# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import os
import httpx
import ssl
from base64 import b64encode
from hashlib import md5
from time import time, sleep, strptime
from calendar import timegm
from threading import Thread
from copy import deepcopy
from logging import getLogger

from . import qtemplate
from ..db import CsvHandler
from ..db import DbHandler


class HttpError(Exception):
    """
    Class for HTTP connection exceptions.
    """


class MqeQuery(Thread):
    """
    A class that handles queries to the HCP MQE API.
    """

    def __init__(self, window, conf, cancel, clearfile, *args, **kwargs):
        """
        :param window:  the calling pySimpleGui window
        :param conf:    a dict of configuration values
        :param cancel:  an Event object, used to signal a stop request
        :param clearfile:  clear the output file if True
        """
        self.log = getLogger('hcpmqe.' + __name__)

        super().__init__(*args, **kwargs)
        self.window = window
        self.conf = conf
        self.cancel = cancel  # used to cancel the query
        self.clearfile = clearfile
        # exposed variables, can be changed during a run
        self.__pagesize = self.conf['count']
        self.__throttle = self.conf['throttle']

        self.qdict = deepcopy(qtemplate)
        self.count = 0

    @property
    def pagesize(self):
        return self.__pagesize

    @pagesize.setter
    def pagesize(self, value):
        self.__pagesize = int(value)

    @property
    def throttle(self):
        return self.__throttle

    @throttle.setter
    def throttle(self, value):
        self.__throttle = int(value)

    def run(self):
        """
        Run the query.
        """
        if self.conf['dbformat'] == 'sqlite3':
            self.db = DbHandler(self.conf['dbfile'], self.clearfile)
        else:
            self.db = CsvHandler(self.conf['dbfile'], self.clearfile, compression=self.conf['compression'])

        # setup a non-verifying SSL-context
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        context.verify_mode = ssl.CERT_NONE
        context.check_hostname = False

        self.ses = httpx.Client(base_url='https://' + self.conf['url'],
                                verify=context,
                                timeout=60.0,
                                headers={
                                    'Authorization': f'HCP {b64encode(self.conf["user"].encode()).decode()}:'
                                                     f'{md5(self.conf["password"].encode()).hexdigest()}',
                                    'Accept': 'application/json',
                                    'Content-Type': 'application/json'}
                                )

        if self.conf['lastResult']['urlName'] and self.conf['lastResult']['changeTimeMilliseconds'] and \
                self.conf['lastResult']['version']:
            lastResult = self.conf['lastResult']
        else:
            lastResult = {}
        first = True
        while True:
            if self.cancel.is_set():
                self.db.close()
                try:  # this will cause a runtime error when the thread ends too quickly, thus ignored
                    self.window.write_event_value('-QTHREAD-ERROR-', ('Query was canceled',))
                except:
                    pass
                return

            try:
                count, urlName, changeTimeMilliseconds, version, finished, qduration, dbduration = self.__query(
                    lastResult=lastResult)
            except HttpError as e:
                self.db.close()
                try:  # this will cause a runtime error when the thread ends too quickly, thus ignored
                    self.window.write_event_value('-QTHREAD-ERROR-', (f'{e}',))
                except:
                    pass
                return

            first = first if not first else False

            if not finished:
                # update window with the latest stats...
                self.window.write_event_value('-QTHREAD-', (count, urlName, version, changeTimeMilliseconds, finished))

                lastResult = {'urlName': urlName,
                              'changeTimeMilliseconds': changeTimeMilliseconds,
                              'version': version
                              }
                if self.__throttle:
                    self.window.write_event_value('-QTHREAD-STATUS-', (f'DB insert took {dbduration:.2f} seconds '
                                                                       f'- now throttling for {self.__throttle} seconds...',
                                                                       'db', dbduration))
                    sleep(self.__throttle)
                self.window.write_event_value('-QTHREAD-STATUS-', (f'DB insert took {dbduration:.2f} seconds - '
                                                                   f'now continue with query for next page...',
                                                                   'db', dbduration))

            else:
                self.db.close()
                # signal window that we are finished...
                try:  # this will cause a runtime error when the thread ends too quickly, thus ignored
                    self.window.write_event_value('-QTHREAD-', (count, urlName, version, changeTimeMilliseconds, finished))
                except:
                    pass
                return

    def __query(self, lastResult={}):
        """
        Make a query against the MQE API.

        :param lastResults: a dict holding values for the next page
        """
        # self.qdict['operation']['count'] = self.conf['count']
        self.qdict['operation']['count'] = self.__pagesize
        if lastResult:
            self.qdict['operation']['lastResult'] = {'urlName': lastResult['urlName'],
                                                     'changeTimeMilliseconds': lastResult['changeTimeMilliseconds'],
                                                     'version': lastResult['version']}
        try:
            __stime = int(timegm(strptime(self.conf['starttime'], '%Y/%m/%d %H:%M:%S')))
        except ValueError as e:
            self.log.debug(f'fatal: invalid Start time - {e}')
            raise HttpError(f'fatal: invalid Start time - {e}')

        self.qdict['operation']['systemMetadata']['changeTime']['start'] = __stime if __stime >= 0 else 0
        try:
            __etime = int(timegm(strptime(self.conf['endtime'], '%Y/%m/%d %H:%M:%S')))
        except ValueError as e:
            self.log.debug(f'fatal: invalid End time - {e}')
            raise HttpError(f'fatal: invalid End time - {e}')

        self.qdict['operation']['systemMetadata']['changeTime']['end'] = __etime * 1000
        self.qdict['operation']['systemMetadata']['directories']['directory'] = \
            self.conf['directories'].split(',') if self.conf['directories'] else []
        self.qdict['operation']['systemMetadata']['namespaces']['namespace'] = \
            self.conf['namespaces'].split(',') if self.conf['namespaces'] else []
        self.qdict['operation']['systemMetadata']['transactions']['transaction'] = self.conf['transactions']
        self.qdict['operation']['verbose'] = self.conf['verbose']

        qduration = 0
        try:
            self.log.debug('now sending query')
            qduration = time()
            r = self.ses.post('/query', json=self.qdict)
            qduration = time() - qduration
        except Exception as e:
            raise HttpError(f'fatal: query failed ({e})')
        else:
            if r.status_code == 403:
                self.log.debug(f'fatal: {r.status_code} - {r.reason_phrase}')
                raise HttpError(f'fatal: {r.status_code} - {r.reason_phrase}')
            elif r.status_code != 200:
                self.log.debug(f'fatal: {r.status_code} - {r.headers["X-HCP-ErrorMessage"]}')
                raise HttpError(f'fatal: {r.status_code} - {r.headers["X-HCP-ErrorMessage"]}')
            else:
                self.log.debug(f'status_code = {r.status_code}')
                res = r.json()
                # print(res)
                # {'queryResult': {'query': {'end': 1603745726000, 'start': 0},
                #                  'resultSet': [{'accessTime': 1600781077,
                #                                 'accessTimeString': '2020-09-22T15:24:37+0200',
                #                                 'acl': False,
                #                                 'changeTimeMilliseconds': '1600781077330.00',
                #                                 'changeTimeString': '2020-09-22T15:24:37+0200',
                #                                 'customMetadata': False,
                #                                 'customMetadataAnnotation': '',
                #                                 'dpl': 1,
                #                                 'gid': 0,
                #                                 'hash': 'MD5 04CBFB438DE89F42FD8C2B426FF44E40',
                #                                 'hashScheme': 'MD5',
                #                                 'hold': False,
                #                                 'index': True,
                #                                 'ingestTime': 1600781077,
                #                                 'ingestTimeString': '2020-09-22T15:24:37+0200',
                #                                 'namespace': 'p-demo.s3',
                #                                 'objectPath': '/Pictures/x1b1YDoKapI.jpg',
                #                                 'operation': 'CREATED',
                #                                 'owner': 'USER,s3,s3user',
                #                                 'permissions': 555,
                #                                 'replicated': True,
                #                                 'replicationCollision': False,
                #                                 'retention': 0,
                #                                 'retentionClass': '',
                #                                 'retentionString': 'Deletion Allowed',
                #                                 'shred': False,
                #                                 'size': 252473,
                #                                 'type': 'object',
                #                                 'uid': 0,
                #                                 'updateTime': 1600781077,
                #                                 'updateTimeString': '2020-09-22T15:24:37+0200',
                #                                 'urlName': 'https://p-demo.s3.hcp80.archivas.com/rest/Pictures/x1b1YDoKapI.jpg',
                #                                 'utf8Name': 'x1b1YDoKapI.jpg',
                #                                 'version': 102449988933185}],
                #                  'status': {'code': 'INCOMPLETE', 'message': '', 'results': 1}}}

                self.window.write_event_value('-QTHREAD-STATUS-',
                                              (f'query took {qduration:.2f} seconds - now inserting '
                                               f'{res["queryResult"]["status"]["results"]:,} records into database',
                                               'query', qduration))
                dbduration = time()
                try:
                    self.db.writeops(res["queryResult"]["resultSet"])
                except Exception as e:
                    raise HttpError(f'fatal: writing to DB file went wrong - {e}')
                dbduration = time() - dbduration

                self.count += res["queryResult"]["status"]["results"]

        if res['queryResult']['resultSet']:
            return self.count, \
                   res['queryResult']['resultSet'][-1]['urlName'], \
                   res['queryResult']['resultSet'][-1]['changeTimeMilliseconds'], \
                   res['queryResult']['resultSet'][-1]['version'], \
                   True if res["queryResult"]["status"]["code"] == 'COMPLETE' else False, \
                   qduration, dbduration

        return self.count, '', '', '', \
               True if res["queryResult"]["status"]["code"] == 'COMPLETE' else False, qduration, dbduration



# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import csv
import bz2
import gzip
import lzma
from logging import getLogger


# noinspection SqlNoDataSourceInspection
class CsvHandler():
    """
    Setup and write a SQLite3 database file.
    """

    def __init__(self, filename, clearfile, compression='plain'):
        """
        :param filename:     the name of the sqlite3 database
        :param compression:  the compression method
        :param clearfile:    clear the output file if True
        """
        self.filename = filename
        self.clearfile = clearfile

        self.log = getLogger('hcpmqe.' + __name__)

        # delete csv file if existent
        if self.clearfile:
            try:
                os.remove(filename)
                self.log.debug(f'deleted existing csv file "{filename}"')
            except (OSError) as e:
                self.log.debug(f'failed to delete csv file "{filename}" - {e}')

        compression = compression.strip("('").strip("',)")

        ## setup csv writer
        try:
            if compression == 'plain':
                self.filehdl = open(self.filename, 'a', newline='')
            elif compression == 'bz2':
                self.filehdl = bz2.open(self.filename, 'at', newline='', compresslevel=9)  # bzip2 -d -c test.csv.bz2
            elif compression == 'gzip':
                self.filehdl = gzip.open(self.filename, 'at', newline='', compresslevel=9)  # gunzip < test.csv.gzip
            elif compression == 'lzma':
                self.filehdl = lzma.open(self.filename, 'at', newline='')  # zcat < test.csv.lzma
            else:
                self.log.debug(f'unable to create csv file - {compression} compression unavailable')
                raise AttributeError(f'unable to create csv file - {compression} compression unavailable')
            self.log.debug(f'created csv file "{filename}" ({compression})')
        except Exception as e:
            self.log.debug(f'failed to create csv file "{filename}" - {e}')
            raise (e)

        self.opscreated = False  # signals that the header is not yet created

    def init_db(self):
        """
        This initializes the csv file - a dummy method.
        """
        self.log.debug('initializing the csv file')


    def __init_ops(self, rec):
        """
        Create the ops table, depending on the records received
        """
        self.csv = csv.DictWriter(self.filehdl, rec.keys())
        self.csv.writeheader()
        self.log.debug(f'wrote csv header -> {list(rec.keys())}')


    def writeops(self, resultset):
        """
        Ingest all records in resultset into DB.
        """
        if not self.opscreated:  # create the ops table based on the first record
            self.__init_ops(resultset[0])
            self.opscreated = True

        self.csv.writerows(resultset)

        self.log.debug(f'wrote {len(resultset)} records into csv file')

    def close(self):
        self.filehdl.close()
        self.log.debug('closed csv file')



# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
from os import getcwd
from pathlib import PurePath
import PySimpleGUI as sg

def buildlayout(conf: dict, readfromdisk: bool = False) -> list:
    """
    Build the GUI layout.

    :param conf:          the configuration dict
    :param readfromdisk:  True if config was successfully read from disk
    :returns:             a list holding the layout
    """
    sg.theme('SystemDefault1')

    # ------ Menu Definition ------ #
    menu_def = [['&File', ['&Load configuration...',
                           '&Save configuration as...',
                           '---',
                           '&Clear configuration',
                           '---',
                           'E&xit']],
                ['&Help', ['&About...',
                           '&Open help in browser...']]
                ]

    # ------ left column ------ #
    HCPacc_l = [[sg.Text('HCP FQDN:'),
                 sg.InputText(conf['url'], key='-URL-', size=(49, 1))],
                [sg.Text('Namespace(s):'),
                 sg.InputText(conf['namespaces'] if conf['namespaces'] else '', key='-NAMESPACES-',
                              size=(49,1))],
                [sg.Text('Directories:'),
                 sg.InputText(conf['directories'] if conf['directories'] else '', key='-DIRECTORIES-',
                              size=(49,1))],
                [sg.Text('User:'),
                 sg.InputText(conf['user'], key='-USER-', size=(49,1))],
                [sg.Text('Password:'),
                 sg.InputText(password_char='*', key='-PASSWORD-', size=(49,1))],
                ]

    HCPload_l = [[sg.Text('Records per page:'),
                  sg.Slider(range=(1, 10000),
                            default_value=int(conf['count']),
                            key='-COUNT-',
                            size=(27, 15),
                            orientation='horizontal'),
                  sg.Button('Set', key='-SET-PAGE-', disabled=True)],
                 [sg.Text('Throttle (sec between pages):'),
                  sg.Slider(range=(0, 60),
                            default_value=int(conf['throttle']),
                            key='-THROTTLE-',
                            size=(27, 15),
                            orientation='horizontal'),
                  sg.Button('Set', key='-SET-THROTTLE-', disabled=True)]
                 ]

    Output_l = [[sg.Text('Output type:'),
                 sg.Radio('csv', 'R-1', key='-CSV-',
                          default=True if conf['dbformat']=='csv' else False),
                 sg.Spin(['plain','bz2','gzip','lzma'], initial_value=[conf['compression']],
                            size=(6,1), key='-COMPRESSION-'),
                 sg.VerticalSeparator(color='dark grey'),
                 sg.Radio('sqlite3', 'R-1', key='-SQLITE3-',
                          default=True if conf['dbformat']=='sqlite3' else False),
                 sg.VerticalSeparator(color='dark grey'),
                 sg.FileSaveAs('Browse', key='-SAVEAS-', target='-FILE-', initial_folder=getcwd())],
                 # sg.FileBrowse('Browse', key='-SAVEAS-', target='-FILE-', initial_folder=getcwd())],
                [sg.Text('Output file:'),
                 sg.InputText(conf['dbfile'] if conf['dbfile'] else '', key='-FILE-', enable_events=True, disabled=True),
                 sg.Checkbox('verbose', key='-VERBOSE-', default=conf['verbose'])],
                ]

    leftupper_l = [[sg.Frame('HCP access parameters:', layout=HCPacc_l, element_justification='right')],
                   [sg.Frame('HCP load parameters:', layout=HCPload_l, element_justification='right')],
                   [sg.Frame('Output parameters:', layout=Output_l, element_justification='right')]
                   ]

    # ------ right column ------ #
    qParams_l = [[sg.Checkbox('create', default=True if 'CREATE' in conf['transactions'] else False,
                              key='-TP-CREATE-')],
                 [sg.Checkbox('delete', default=True if 'DELETE' in conf['transactions'] else False,
                              key='-TP-DELETE-')],
                 [sg.Checkbox('dispose', default=True if 'DISPOSE' in conf['transactions'] else False,
                              key='-TP-DISPOSE-')],
                 [sg.Checkbox('prune', default=True if 'PRUNE' in conf['transactions'] else False,
                              key='-TP-PRUNE-')],
                 [sg.Checkbox('purge', default=True if 'PURGE' in conf['transactions'] else False,
                              key='-TP-PURGE-')],
                 ]

    timeRange_l = [[sg.Text('Start time:'),
                    sg.InputText(conf['starttime'], key='-STARTTIME-', size=(23,1)),
                    sg.Button('Reset', key='-RESET-START-')],
                   [sg.Text('End time:'),
                    sg.InputText(conf['endtime'], key='-ENDTIME-', size=(23,1)),
                    sg.Button('Reset', key='-RESET-END-')],
                   ]

    rightupper_l = [[sg.Frame('Transaction types:', layout=qParams_l)],
                    [sg.Frame('Time range:', layout=timeRange_l, element_justification='right')],
                    ]

    baseframe_t = [[sg.Text('Status:'),
                    sg.Text('settings were successfully read from history file' if readfromdisk else
                            'unable to read settings from history file', key='-STATUS-', size=(100,1))],
                   [sg.Text('Records found:'),
                    sg.Text('', key='-RECS-FOUND-', size=(100,1))]]

    baseframe_b = [
                   # [sg.Text('Last record:', justification='top'),
                   #  sg.Text('', size=(100, 1), justification='left')],
                   [sg.Text('Url:', justification='top'),
                    sg.InputText(conf['lastResult']['urlName'] if 'urlName' in conf['lastResult'] else '',
                                 key='-LAST-URL-', size=(110, 1), justification='left')],
                   [sg.Text('Version:', justification='top'),
                    sg.InputText(conf['lastResult']['version'] if 'version' in conf['lastResult'] else '',
                                 key='-LAST-VERSION-', size=(110, 1), justification='left')],
                   [sg.Text('ChgTimeSec:', justification='top'),
                    sg.InputText(conf['lastResult']['changeTimeMilliseconds'] if 'changeTimeMilliseconds' in conf['lastResult'] else '',
                                 key='-LAST-CHGTIMEMS-', size=(110, 1), justification='left')]]

    layout = [[sg.Menu(menu_def)],
              [sg.Frame('', layout=leftupper_l, border_width=0, element_justification='right'),
               sg.VerticalSeparator(color='dark grey'), sg.Frame('Query parameters', layout=rightupper_l, vertical_alignment='top', border_width=0)],
              [sg.Frame('', layout=baseframe_t, border_width=0, element_justification='right', vertical_alignment='top')],
              [sg.Frame('Last record', layout=baseframe_b, border_width=1, element_justification='right', vertical_alignment='top')],
              [sg.Frame('',
                        [[sg.Submit('Run query', key='-RUN-QUERY-', bind_return_key=False),
                          sg.Cancel('Quit', key='-QUIT-'),
                          sg.HorizontalSeparator(),
                          sg.Text('',size=(100,1),key='-TIMESTAMP-')]],
                        vertical_alignment='right', element_justification='right', border_width=0)]]

    return layout

def updatelayout(window, conf):
    """
    Update the GUI with values from the config.

    :param window:        the window object
    :param conf:          the configuration dict
    :param readfromdisk:  True if config was successfully read from disk
    """
    window['-URL-'].update(conf['url'])
    window['-USER-'].update(conf['user'])
    window['-PASSWORD-'].update('')
    window['-COUNT-'].update(conf['count'])
    window['-THROTTLE-'].update(conf['throttle'])
    window['-NAMESPACES-'].update(conf['namespaces'] if conf['namespaces'] else '')
    window['-DIRECTORIES-'].update(conf['directories'] if conf['directories'] else '')

    window['-TP-CREATE-'].update(value=True if 'CREATE' in conf['transactions'] else False)
    window['-TP-DELETE-'].update(value=True if 'DELETE' in conf['transactions'] else False)
    window['-TP-DISPOSE-'].update(value=True if 'DISPOSE' in conf['transactions'] else False)
    window['-TP-PRUNE-'].update(value=True if 'PRUNE' in conf['transactions'] else False)
    window['-TP-PURGE-'].update(value=True if 'PURGE' in conf['transactions'] else False)

    window['-STARTTIME-'].update(conf['starttime'])
    window['-ENDTIME-'].update(conf['endtime'])
    window['-VERBOSE-'].update(conf['verbose'])
    window['-CSV-'].update(True if conf['dbformat']=='csv' else False)
    window['-SQLITE3-'].update(True if conf['dbformat']=='sqlite3' else False)
    window['-COMPRESSION-'].update(conf['compression'])
    window['-FILE-'].update(conf['dbfile'])

    if conf['lastResult'] and conf['lastResult']['urlName'] and conf['lastResult']['changeTimeMilliseconds'] and conf['lastResult']['version']:
        window['-LAST-URL-'].update(conf['lastResult']['urlName'])
        window['-LAST-CHGTIMEMS-'].update(conf['lastResult']['changeTimeMilliseconds'])
        window['-LAST-VERSION-'].update(conf['lastResult']['version'])
    else:
        window['-LAST-URL-'].update('')
        window['-LAST-CHGTIMEMS-'].update('')
        window['-LAST-VERSION-'].update('')



# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import sqlite3
from time import strftime, localtime, time
from logging import getLogger

from ..version import Gvars


# noinspection SqlNoDataSourceInspection
class DbHandler():
    """
    Setup and write a SQLite3 database file.
    """

    def __init__(self, filename, clearfile):
        """
        :param filename:   the name of the sqlite3 database
        :param clearfile:  clear the output file if True
        """
        self.filename = filename
        self.clearfile = clearfile

        self.log = getLogger('hcpmqe.' + __name__)

        # delete database file if existant
        if self.clearfile:
            try:
                os.remove(filename)
                self.log.debug(f'deleted existing db file "{filename}"')
            except (OSError) as e:
                self.log.debug(f'failed to delete db file "{filename}" - {e}')
                pass

        ## setup database
        try:
            self.dbConn = sqlite3.connect(self.filename)
            self.dbConn.isolation_level = None  # use AUTOCOMMIT mode
            self.log.debug(f'connected to DB file "{filename}"')
        except sqlite3.DatabaseError as e:
            self.log.debug(f'failed to connect to DB file "{filename}" - {e}')
            raise (e)

        self.opscreated = False  # signals that the ops table is not yet created
        self.ic = None

    def __init_db(self):
        """
        This initializes the database with the proper tables.
        """
        self.log.debug('initializing the database')
        self.ic = self.dbConn.cursor()

        dbSchema = {"admin": '''CREATE TABLE admin
                                (magic           TEXT,
                                 version         TEXT,
                                 creation        TEXT)'''
                    }

        # initialize the needed database tables
        for k in dbSchema.keys():
            self.ic.execute(dbSchema[k])

        # write admin record
        self.ic.execute('INSERT INTO admin VALUES (?,?,?)', ('hcpmqt', Gvars.Version,
                                                             strftime("%a, %Y/%m/%d, %H:%M:%S",
                                                                      localtime(time()))))

        self.dbConn.commit()

    def __init_ops(self, fullinit, rec):
        """
        Create the ops table, depending on the records received

        :param fullinit:  create the table if True
        :param rec:       the first record to carve the columns from
        """
        if fullinit:
            schema = f'CREATE TABLE ops ({", ".join(rec.keys()).replace("index", "_index")})'
            self.log.debug(f'Schema -> {schema}')
            self.ic.execute(schema)
            self.dbConn.commit()

        self.insert = f'INSERT INTO ops VALUES ({", ".join([f":{x}" for x in rec.keys()])})'
        self.log.debug(f'self.insert -> {self.insert}')

    def writeops(self, resultset):
        """
        Ingest all records in resultset into DB.
        """
        if not self.opscreated:  # in case this is the first record...
            if self.clearfile:   # ...create the ops table based on the first record
                self.__init_db()
            self.opscreated = True
            try:
                self.__init_ops(self.clearfile, resultset[0])
            except IndexError as e:
                self.log.debug(f'resultset was empty - inserted {len(resultset)} records into db')
                return

        if not self.ic:
            self.ic = self.dbConn.cursor()

        # Using a transaction speeds up DB inserts drastically, as it turns off auto-commit
        # of every single INSERT, in exchange for just one commit for the while batch.
        # In addition, it allows for a clean continuation in case the query got stuck due
        # to an error.
        self.ic.execute('BEGIN IMMEDIATE TRANSACTION')
        self.ic.executemany(self.insert, resultset)

        self.dbConn.commit()
        self.log.debug(f'inserted {len(resultset)} records into db')

    def close(self):
        self.dbConn.close()
        self.log.debug('closed db')



# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import os


def logger(logtarget='console', level='INFO'):
    """
    Setup loggging, either to journald or to file.

    :param logtarget:  where to log: console, logfile or journald
    :param level:      the log level (INFO or DEBUG)
    """
    logtarget = logtarget.upper()
    level = level.upper()

    if logtarget not in ['CONSOLE', 'LOGFILE']:
        raise AttributeError(f'unable to log to "{logtarget}"')
    if level not in ['INFO', 'DEBUG']:
        raise AttributeError(f'log level "{level}" is not supported')

    log = logging.getLogger('hcpmqe')
    log.setLevel(logging.DEBUG if level=='DEBUG' else logging.INFO)

    if logtarget == 'CONSOLE':
        th = logging.StreamHandler()
        if level == 'DEBUG':
            fm = logging.Formatter('%(asctime)s [%(levelname)s] %(name)s  %(lineno)s : %(message)s')
        else:
            fm = logging.Formatter('[%(levelname)s] : %(message)s')
    else:  # LOGFILE
        logfile = f'hcpmqe.{os.getpid()}.log'

        # delete logfile if existent
        try:
            os.remove(logfile)
        except (OSError) as e:
            pass

        th = logging.FileHandler(logfile)
        if level == 'DEBUG':
            fm = logging.Formatter('%(asctime)s [%(levelname)s] %(name)s : %(message)s')
        else:
            fm = logging.Formatter('%(asctime)s [%(levelname)s] : %(message)s')

    log.addHandler(th)
    th.setFormatter(fm)

    return log


# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

qtemplate = {'operation': {'count': 'num-recs-requested',
                           # commented out, as it is not needed in the 1st request
                           # 'lastResult': {'urlName': 'object-url',
                           #                'changeTimeMilliseconds': 'change-time-in-milliseconds.index',
                           #                'version': 'version-id'
                           #                },
                           # 'objectProperties': 'comma-separated-list-of-properties',
                           'systemMetadata': {'changeTime': {'start': 'start-time-in-milliseconds',
                                                             'end': 'end-time-in-milliseconds'
                                                             },
                                              'directories': {'directory': ['directory-path', '...']
                                                              },
                                              # 'indexable': 'bool',
                                              'namespaces': {'namespace': ['namespace-name.tenant-name', '...']
                                                             },
                                              # 'replicationCollision': 'bool',
                                              'transactions': {
                                                  'transaction': ['operation-type', '...']
                                              }
                                              },
                           'verbose': 'bool'
                           }
             }




# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from os import getcwd
from time import localtime, strftime, time
from threading import Event
from logging import getLogger
from pathlib import PurePath
import PySimpleGUI as sg

from ..conf.config import Configuration
from .layout import buildlayout, updatelayout
from ..mqe import MqeQuery
from ..version import Gvars


def loop():
    log = getLogger('hcpmqe.' + __name__)

    # read the configuration file and build the layout
    conf = Configuration()
    c, r = conf.readconfig()
    layout = buildlayout(c, r)
    title = f'HCP Metadata Query Tool v{Gvars.s_version}'
    sg.theme('SystemDefault1')

    window = sg.Window(title, layout,
                       default_element_size=(40, 1),
                       grab_anywhere=False)

    configpath = getcwd()
    configfile = None
    t_run = t_query = t_db = 0  # used to record the seconds consumed
    mq = None  # placeholder for the query class

    while True:
        event, values = window.read()
        # update the bottom status line while a query thread is running
        if t_run:
            tstxt = f'run time: {hms(time()-t_run)}, ' \
                    f'overall query time: {hms(t_query)}, ' \
                    f'overall DB insert time: {hms(t_db)}'
            window['-TIMESTAMP-'].update(f'{tstxt:>130}')

        # ----- USER ACTIVITY ----- #
        # handle the user closing the window (hard program abort)
        if not event:
            log.debug("user closed the window (abort)")
            break

        # handle the user ending the program cleanly
        elif event == '-QUIT-' or event == 'Exit':
            log.debug("user selected the [Quit] button or [Exit] from menu")
            if configfile:
                conf.updateconf(values)
                if conf.writeconfig(configfile):
                    log.debug(f'auto-updated configuration file {configfile}')
                else:
                    log.debug(f'fatal: failed to auto-update configuration file {configfile}')

            break

        # handle menu:file/load...
        elif event == 'Load configuration...':
            log.debug(f'user selected "{event}" from the menu')
            configfile = sg.popup_get_file('Load configuration file',
                                           title='Load configuration file...',
                                           default_path=configfile or '',
                                           default_extension='.mqe',
                                           initial_folder=configpath,
                                           save_as=False)
            _c, _r = conf.readconfig(configfile)
            if _r:
                updatelayout(window, _c)
                window['-STATUS-'].update(f'loaded configuration file {configfile}')
                window.set_title(f'{title} - {PurePath(configfile).name}')
            else:
                configfile = None

        # handle menu:file/save...
        elif event == 'Save configuration as...':
            log.debug(f'user selected "{event}" from the menu')
            configfile = sg.popup_get_file('Save configuration file',
                                           title='Save configuration file as...',
                                           default_path=configfile or '',
                                           default_extension='.mqe',
                                           initial_folder=configpath,
                                           save_as=True)
            if configfile:
                conf.updateconf(values)
                if conf.writeconfig(configfile):
                    window['-STATUS-'].update(f'configuration´written to file {configfile}')
                    window.set_title(f'{title} - {PurePath(configfile).name}')
                else:
                    window['-STATUS-'].update(f'fatal: failed to write configuration to file {configfile}')
                    configfile = None

        # handle menu:file/clear
        elif event == 'Clear configuration':
            log.debug(f'user selected "{event}" from the menu')
            if sg.popup_yes_no('Do you really want to clear the configuration?') == 'Yes':
                log.debug(f'user decided to clear the configuration')
                configfile = None
                _c, _r = conf.readconfig(configfile)
                updatelayout(window, _c)
                window['-STATUS-'].update(f'configuration cleared')
                window.set_title(title)
            else:
                og.debug(f'user decided NOT to clear the configuration')

        # handle menu:help/about...
        elif event == 'About...':
            sg.popup_ok(f'{Gvars.s_description}\n'
                        f'{Gvars.Version}\n\n'
                        f'Operational queries towards HCP MQE\n\n'
                        f'{Gvars.Author}\n'
                        f'{Gvars.AuthorMail}',
                        background_color='khaki')

        # handle menu:help/Open help in browser...
        elif event == 'Open help in browser...':
            log.debug("user selected [Help] in menu")
            import webbrowser
            webbrowser.open_new_tab(Gvars.AppURL)

        # handle the user changing the query page size
        elif event == '-SET-PAGE-':
            log.debug("user selected the [SET] Records per page button")
            mq.pagesize = values['-COUNT-']

        # handle the user changing the throttle setting
        elif event == '-SET-THROTTLE-':
            log.debug("user selected the [SET] Throttle button")
            mq.throttle = values['-THROTTLE-']

        # handle the user resetting the start time
        elif event == '-RESET-START-':
            log.debug("user selected the [Reset] start time button")
            window['-STARTTIME-'].update('1970/01/01 00:00:00')

        # handle the user resetting the end time
        elif event == '-RESET-END-':
            log.debug("user selected the [Reset] end time button")
            window['-ENDTIME-'].update(strftime('%Y/%m/%d %H:%M:%S', localtime()))

        # handle the user starting or canceling the query
        elif event == '-RUN-QUERY-':
            if window['-RUN-QUERY-'].GetText() == 'Run query':  # start query
                log.debug("user selected the [Run query] button")
                if not configfile:
                    log.debug("unable to run a query, as no config file was available")
                    sg.popup_error('Please save the configuration before running a query...')
                    continue

                if not values["-FILE-"]:
                    log.debug("unable to run a query, as no output file was selected")
                    sg.popup_error('No output file selected...')
                else:
                    if values['-LAST-URL-'] and values['-LAST-VERSION-'] and values['-LAST-CHGTIMEMS-']:
                        continuequery = sg.popup_yes_no('Do you want to continue the query from the position '
                                                        'shown as last record?\n\n WARNING: this will lead to false '
                                                        'results in case ANY parameter has been changed!!!',
                                                        text_color='red2',
                                                        # background_color='khaki',
                                                        line_width=100, keep_on_top=True)
                        if continuequery == 'No':
                            log.debug("user selected to start with a new query")
                        else:
                            log.debug("user selected to continue from the values in last record")

                    if not values['-LAST-URL-'] or continuequery == 'No':
                        if sg.popup_yes_no(f'Running the query will overwrite the file\n'
                                         f'{values["-FILE-"]}\n'
                                         f'Do you really want to do that?',
                                         text_color='red2',
                                         # background_color='khaki',
                                         line_width=100,
                                         keep_on_top=True) == 'Yes':
                            values['-LAST-URL-'] = ''
                            values['-LAST-VERSION-'] = ''
                            values['-LAST-CHGTIMEMS-'] = ''
                            window['-LAST-URL-'].update('')
                            window['-LAST-VERSION-'].update('')
                            window['-LAST-CHGTIMEMS-'].update('')
                            log.debug("user selected to overwrite the db file")
                        else:
                            log.debug("user selected not to run the query")
                            continue

                    togglequery(window, values, True)
                    window['-STATUS-'].update('Query started...')
                    window['-RECS-FOUND-'].update('')
                    window['-TIMESTAMP-'].update('')
                    if configfile:
                        conf.updateconf(values)
                        if conf.writeconfig(configfile):
                            log.debug(f'auto-updated configuration file {configfile} (query start)')
                        else:
                            log.debug(f'fatal: failed to auto-update configuration file {configfile} (query start)')

                    cancel = Event()  # used to cancel a running query
                    mq = MqeQuery(window, conf.conf, cancel, False if values['-LAST-URL-'] else True,
                                  daemon=True, name='*query*')
                    t_run = time()
                    t_query = t_db = 0
                    mq.start()
            else:  # cancel query
                log.debug("user selected the [Cancel query] button")
                if sg.popup_yes_no('Do you really want to cancel the running query?', text_color='red2',
                                   # background_color='khaki'
                                   ) == 'Yes':
                    log.debug("user chose to cancel the query")
                    cancel.set()
                    window['-RUN-QUERY-'].update(disabled=True)
                    window['-STATUS-'].update('Query is being canceled - please wait for finishing')
                    window.Finalize()
                else:
                    log.debug("user chose NOT to cancel the query")

        # handle the user changing the target file
        # elif event == '-xFILE-':
        #     if values['-FILE-']:
        #         _c, __dummy = conf.readconfig(values['-FILE-'])
        #         updatelayout(window, _c)
        #         window['-STATUS-'].update(f'read configuration from {PurePath(values["-FILE-"]).with_suffix(".mqeconf")}')
        #         window.Finalize()

        # ----- QUERY THREAD ACTIVITY ----- #
        # handle messages from the query thread
        elif event == '-QTHREAD-' and len(values['-QTHREAD-']) > 1:  # the query didn't fail
            if values['-QTHREAD-'][4]:  # finished
                log.debug("Query thread signaled query has finished successfully")
                mq.join()
                mq = None
                window['-RECS-FOUND-'].update(values['-QTHREAD-'][0])
                window['-LAST-URL-'].update(values["-QTHREAD-"][1]),
                window['-LAST-VERSION-'].update(values["-QTHREAD-"][2]),
                window['-LAST-CHGTIMEMS-'].update(values["-QTHREAD-"][3]),
                window['-STATUS-'].update('finished')
                togglequery(window, values, False)
            else:  # not finished, ongoing...
                log.debug("Query thread signaled query information update")
                window['-RECS-FOUND-'].update(values['-QTHREAD-'][0])
                window['-LAST-URL-'].update(values["-QTHREAD-"][1]),
                window['-LAST-VERSION-'].update(values["-QTHREAD-"][2]),
                window['-LAST-CHGTIMEMS-'].update(values["-QTHREAD-"][3]),
                window['-STATUS-'].update(values["-QTHREAD-"][0])

            # need to tweak the machinery here, to get these values updated
            values['-LAST-URL-'] = values["-QTHREAD-"][1]
            values['-LAST-VERSION-'] = values["-QTHREAD-"][2]
            values['-LAST-CHGTIMEMS-'] = values["-QTHREAD-"][3]
            conf.updateconf(values)
            if conf.writeconfig(configfile):
                log.debug(f'auto-updated configuration file {configfile} (status update)')
            else:
                log.debug(f'fatal: failed to auto-update configuration file {configfile} (status update)')


        # handle status updated submitted by the query thread
        elif event == '-QTHREAD-STATUS-':
            log.debug(f'Query thread signaled status update ({values["-QTHREAD-STATUS-"][0]})')
            window['-STATUS-'].update(values['-QTHREAD-STATUS-'][0])
            if values["-QTHREAD-STATUS-"][1] == 'db':
                t_db += float(values["-QTHREAD-STATUS-"][2])
            elif values["-QTHREAD-STATUS-"][1] == 'query':
                t_query += float(values["-QTHREAD-STATUS-"][2])

        # handle the query thread signalling an error that causes the query to end
        elif event == '-QTHREAD-ERROR-':  # the query failed/canceled
            log.debug(f'Query thread signaled status update ({values["-QTHREAD-ERROR-"][0]})')
            window['-STATUS-'].update(values['-QTHREAD-ERROR-'][0])
            togglequery(window, values, False)
            mq.join()
            mq = None

            if values['-QTHREAD-ERROR-'][0].find('403 -') >= 0:
                conf.updateconf(values)
                if conf.writeconfig(configfile):
                    log.debug(f'auto-updated configuration file {configfile} (query cancel/fail)')
                else:
                    log.debug(f'fatal: failed to auto-update configuration file {configfile} (query cancel/fail)')
            else:
                log.debug(f'NOT auto-updating the configuration file {configfile} (due to error 403)')

    window.close()
    log.debug('program ended')


def togglequery(window, values, run: bool):
    """
    Toggle the query status in the GUI.

    :param window:  the window
    :param values:  the values dict()
    :param run:     True if we toggle to run query mode
    """
    # groups of elements to be toggled when running a query / query finished or canceled
    t1 = ['-URL-', '-NAMESPACES-', '-DIRECTORIES-', '-USER-', '-PASSWORD-',
          '-CSV-', '-COMPRESSION-', '-SQLITE3-', '-SAVEAS-',  '-FILE-', '-VERBOSE-',
          '-TP-CREATE-', '-TP-DELETE-', '-TP-DISPOSE-', '-TP-PRUNE-', '-TP-PURGE-',
          '-STARTTIME-', '-RESET-START-',  '-ENDTIME-', '-RESET-END-', '-QUIT-']
    t2 = ['-SET-PAGE-', '-SET-THROTTLE-']

    if run:
        for el in t1:
            window[el].update(disabled=True)
        for el in t2:
            window[el].update(disabled=False)
        window['-RUN-QUERY-'].update(text='Cancel query')
    else:
        for el in t1:
            window[el].update(disabled=False)
        for el in t2:
            window[el].update(disabled=True)
        window['-RUN-QUERY-'].update(text='Run query', disabled=False)


def hms(t):
    """
    Return a sting of HH:MM:SS from an integer
    """
    h = int(t/3600)
    m = int((t-h*3600)/60)
    s = int(t-h*3600-m*60)

    return f'{h:02}:{m:02}:{s:02}'




# The MIT License (MIT)
#
# Copyright (c) 2012-2020 Thorsten Simons (sw@snomis.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
from logging import getLogger
from os.path import expanduser
from pathlib import PurePath
from time import time, strftime, localtime
from copy import deepcopy


class Configuration():
    """
    Read/Write configuration to disk.

    Configuration file is in ~/.hcpmqe.conf, in json format.
    """

    def __init__(self):
        self.log = getLogger('hcpmqe.' + __name__)

    def emptyconfig(self):
        """
        Create an empty configuration structure.
        """
        self.conf = {'url': '',
                     'user': '',
                     'password': '',
                     'count': 2500,
                     'throttle': 0,
                     'namespaces': [],
                     'directories': [],
                     'transactions': ['CREATE'],
                     'starttime': '1970/01/01 00:00:00',
                     'endtime': strftime('%Y/%m/%d %H:%M:%S', localtime()),
                     'verbose': False,
                     'dbformat': 'sqlite3',
                     'compression': 'plain',
                     'dbfile': '',
                     'lastResult': {
                         'urlName': '',
                         'changeTimeMilliseconds': '',
                         'version': ''
                     },
                     }

    def readconfig(self, file=None) -> dict:
        """
        Read the configuration file from a side file.

        :param file:  the name of the data file chosen by the user
        :return:      the configuration dict
        """
        readfromdisk = False
        try:
            with open(file, 'r') as rhdl:
                self.conf = json.load(rhdl)
                readfromdisk = True
                self.log.debug(f'read configuration from {file}')
        except Exception as e:
            self.log.debug(f'unable to read configuration from {file} ({e}) - '
                             f'using default values')
            self.emptyconfig()

        return self.conf, readfromdisk

    def updateconf(self, values):
        """
        Update the conf dict with the windows content.

        :param values:  the pysimpleguy values object
        """
        self.conf['url'] = values['-URL-']
        self.conf['user'] = values['-USER-']
        self.conf['password'] = values['-PASSWORD-']
        self.conf['count'] = int(values['-COUNT-'])
        self.conf['throttle'] = values['-THROTTLE-']
        self.conf['namespaces'] = values['-NAMESPACES-']
        self.conf['directories'] = values['-DIRECTORIES-']
        self.conf['transactions'] = [x[4:-1] for x in values.keys() if str(x).startswith('-TP-') and values[x]]
        self.conf['starttime'] = values['-STARTTIME-']
        self.conf['endtime'] = values['-ENDTIME-']
        self.conf['verbose'] = values['-VERBOSE-']
        self.conf['dbformat'] = 'csv' if values['-CSV-'] else 'sqlite3'
        self.conf['compression'] = values['-COMPRESSION-'].strip("('").strip("',)")
        self.conf['dbfile'] = values['-FILE-']
        if '-LAST-URL-' in values and '-LAST-VERSION-' in values and '-LAST-CHGTIMEMS-' in values:
            self.conf['lastResult'] = {'urlName': values['-LAST-URL-'],
                                       'changeTimeMilliseconds': values['-LAST-CHGTIMEMS-'],
                                       'version': values['-LAST-VERSION-']
                                       }
        else:
            self.conf['lastResult'] = {}

        self.log.debug('updated conf from window entries')

    def writeconfig(self, file) -> bool:
        """
        Write the configuration to a file.

        :param file:  the name of the file chosen by the user
        :returns:    True if successfull, else False
        """
        c = deepcopy(self.conf)
        del c['password']  # we don't want a password to be stored to disk, right?
        try:
            with open(file, 'w') as whdl:
                json.dump(c, whdl, indent=2)
                self.log.debug(f'wrote config to file {file}')
                return True
        except Exception as e:
            self.log.debug(f'unable to write configuration to file {file} ({e}) - skipping')
            return False

